/* ---------------------------------------------------------------------------------------------------------------------------- */
/* CREATE DB SCHEMAS as specified by Table
	FOR SPECIFIED DATABASE*/
/* CREATE THE DATABASE BEFORE EXECUTING THIS SCRIPT */

/* SELECT INTENDED DATABASE, IN SSMS, MSVisualStudio, etc., VIA DROPDOWN IN MENU BAR OR BY SPECIFYING DATABASE WHEN CONNECTING FILE TO RUN  */

/* If an error is encountered at execution of this SQL file, stating that "setproperty must be executed in the UI thread, 
	that can be fixed by repairing OLEDB using its installer. There may be other configuration methods to repair this, 
	but they are currently not identified. */

/* If error is reported in Message stating unable to initialize ACE OLEDB, make sure to close the Excel source Table doc. */

/* ---------------------------------------------------------------------------------------------------------------------------- */
/* Security Architecture:  
	https://learn.microsoft.com/en-us/sql/relational-databases/security/authentication-access/getting-started-with-database-engine-permissions?view=sql-server-ver16
	https://www.mssqltips.com/sqlservertip/2699/sql-server-user-defined-server-roles/ */

/* ---------------------------------------------------------------------------------------------------------------------------- */

/* Entire work Copyright (C) 2022 Richard M. Bixler, All Rights Reserved. Licensing Terms Stated Below. */

/* ---------------------------------------------------------------------------------------------------------------------------- */

/* Extract specifications and sequence to GRANT DB User-Role Permissions */
/* From Table rngOpIUsers specifying DB Users and OpI Role in the selected Database */
SET NOCOUNT ON; 

DECLARE @schLineNo AS INT
DECLARE	@schDatabase AS NVARCHAR(max) = NULL
DECLARE @schScope as NVARCHAR(max) = NULL
DECLARE @schActiveNow AS NVARCHAR(max) = NULL
DECLARE @schSchemaVerb AS NVARCHAR(max) = NULL
DECLARE @schSchemaName AS NVARCHAR(max) = NULL
DECLARE	@schOwnerDomain AS NVARCHAR(max) = NULL
DECLARE	@schOwnerName AS NVARCHAR(max) = NULL
DECLARE @schTransferEntityType AS NVARCHAR(max) = NULL
DECLARE	@schTransferFromSchema AS NVARCHAR(max) = NULL
DECLARE	@schObjectToTransfer AS NVARCHAR(max) = NULL


/* Local Variables */
DECLARE @currentDatabase AS NVARCHAR(max) = NULL
DECLARE @currentUserName as NVARCHAR(max) = NULL
DECLARE	@query AS NVARCHAR(max) = NULL
DECLARE	@DBWrapQuery AS NVARCHAR(max) = NULL
DECLARE @OnObj AS NVARCHAR(max) = NULL
DECLARE @member_principalID AS NVARCHAR(max) = NULL
DECLARE @memberCount AS INT = 0
DECLARE @usrRoleNameFixed AS NVARCHAR(max) = NULL
DECLARE @memberName AS NVARCHAR(max) = NULL
DECLARE @roleName AS NVARCHAR(max) = NULL
DECLARE @createOrAlter AS NVARCHAR(max) = NULL
DECLARE @with AS NVARCHAR(max) = NULL
DECLARE @entityType as NVARCHAR(max) = NULL

/* CURSOR Blocks */
DECLARE @OpIMemberCount AS INT = 0

/* TRY...CATCH Blocks */
DECLARE @Plural as NVARCHAR(1) = N''
DECLARE @errorNumber as INT = 0
DECLARE @errorMessage as NVARCHAR(max) = NULL
DECLARE @errorLine as INT = 0

PRINT N'SECURITY_3-CREATE-DB-Schema.sql'; PRINT N'';

SET @currentDatabase = DB_NAME();	-- Where are we?

DECLARE schemacreate_query_cursor CURSOR FOR 
	SELECT schLineNo, schDatabase, schScope, schActiveNow, schSchemaVerb, schSchemaName, schOwnerDomain, schOwnerName, schTransferEntityType, schTransferFromSchema, schObjectToTransfer
		FROM LS_OpISQLServerSecurity_ToolSource...rngOpISchemaCreate AS xlOpISchemasCreate
		
		WHERE (xlOpISchemasCreate.schScope = 'DB' 
			AND NOT(xlOpISchemasCreate.schActiveNow = 'No')
			AND xlOpISchemasCreate.schSchemaVerb IS NOT NULL
			AND xlOpISchemasCreate.schDatabase = @currentDatabase)

		ORDER BY 
			xlOpISchemasCreate.schLineNo;

OPEN schemacreate_query_cursor;
FETCH NEXT FROM schemacreate_query_cursor 
	INTO @schLineNo, @schDatabase, @schScope, @schActiveNow, @schSchemaVerb, @schSchemaName, @schOwnerDomain, @schOwnerName, @schTransferEntityType, @schTransferFromSchema, @schObjectToTransfer;

WHILE @@FETCH_STATUS = 0 
BEGIN
	IF @schActiveNow IN (N'Yes', N'DryRun')
	BEGIN
		PRINT N'';
		IF @schActiveNow = N'DryRun' PRINT N'Dry Run';
		PRINT CONCAT(N'Row ', @schLineNo, N' ', @schSchemaVerb, N' ', @schSchemaName, N' ', @schDatabase);
		
		IF ( @schSchemaVerb = N'DROP' )
		BEGIN
			/* Drop previously-set OpI DB Schema. */
			IF EXISTS (SELECT [sys].[schemas].[name] 
				FROM [sys].[schemas] 
				WHERE [name] = @schSchemaName)
			BEGIN
				/* DROP this OpI schema. */
				BEGIN
					BEGIN TRY
						SET @errorNumber = 0;
						SET @query = CONCAT(N'DROP SCHEMA ', @schSchemaName, N';');	
						IF @schActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
						SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
				END;
			END;
		END;


		IF ( @schSchemaVerb = N'ADD/UPDATE' )
		BEGIN
			/* Create a clean instance of this OpI DB User if it does not exist */
			IF NOT EXISTS (SELECT [sys].[schemas].[name] 
				FROM [sys].[schemas] 
				WHERE [name] = @schSchemaName)
			BEGIN 
				IF EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
					FROM [sys].[database_principals] 
					WHERE [type] IN (N'U', N'S', N'R') AND [name] = @schOwnerName) 
				BEGIN
					/* OpI DB User Name follows LoginName */
					IF @schOwnerDomain IS NOT NULL SET @schOwnerName = CONCAT(N'[', @schOwnerDomain, N'\', @schOwnerName, N']');
					BEGIN TRY
						SET @errorNumber = 0;
						SET @query = CONCAT(N'CREATE SCHEMA ', @schSchemaName, ' AUTHORIZATION ', @schOwnerName, N';'); 
						IF @schActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
						SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
				END;
				ELSE PRINT CONCAT(N'Schema "', @schSchemaName, '" not created, Principal "', @schOwnerName, N'" does not exist');
			END;
			ELSE 
			BEGIN
				/* Insert ALTER USER code here */
				PRINT CONCAT(N'Schema "', @schSchemaName, N'" already exists');
			END;
		END;


		IF ( @schSchemaVerb = N'ALTER' )
		BEGIN
			SET @with = NULL;
			SET @entityType = NULL;
			IF @schTransferEntityType IS NOT NULL SET @entityType = CONCAT(N' ', @schTransferEntityType, N'::');
			IF ((@schTransferFromSchema IS NOT NULL) AND (@schObjectToTransfer IS NOT NULL)) 
				SET @with = CONCAT(@schTransferEntityType, @schTransferFromSchema, N'.', @schObjectToTransfer);
			SET @with = CONCAT(N' TRANSFER ', @entityType, @with);

			BEGIN TRY
				SET @errorNumber = 0;
				SET @query = CONCAT(N'ALTER SCHEMA ', @schSchemaName, @with, N';'); 
				IF @schActiveNow = N'Yes' EXEC sp_executesql @query;
			END TRY
			BEGIN CATCH
				SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
			END CATCH
			PRINT @query;
			IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
		END;
	END;

	FETCH NEXT FROM schemacreate_query_cursor 
		INTO @schLineNo, @schDatabase, @schScope, @schActiveNow, @schSchemaVerb, @schSchemaName, @schOwnerDomain, @schOwnerName, @schTransferEntityType, @schTransferFromSchema, @schObjectToTransfer;

END;

CLOSE schemacreate_query_cursor;
DEALLOCATE schemacreate_query_cursor;

/* LICENSING
Power Operational Intelligence ("PowerOpI") Framework
   Copyright (C) 2020 Richard M. Bixler
Derived products MSAccessTool, PowerQueryTool, SQLServerTool, and SQLServerRemote
   Copyright � 2021 Richard M. Bixler
Derived product OpISQLServerSecurity
   Copyright � 2022 Richard M. Bixler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or 
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Contact:
rmbpublic@comcast.net
www.softtoyssoftware.com */